import abc

from boto3.dynamodb.conditions import Key

from . import iam
from .exceptions import WecoDeployError


class ReleaseStoreError(WecoDeployError):
    pass


class ReleaseNotFoundError(ReleaseStoreError):
    """
    Raised if a release store is asked to retrieve a release ID which does not exist.
    """

    pass


class ReleaseStore(abc.ABC):
    @abc.abstractmethod
    def describe_initialisation(self):
        """
        Returns a human-readable string to describe the initialisation process.
        """
        pass

    @abc.abstractmethod
    def initialise(self):
        """
        Do any initialisation steps to make the store ready for use.
        """
        pass

    @abc.abstractmethod
    def put_release(self, release):
        """
        Store a new release.  Returns None.
        """
        pass

    @abc.abstractmethod
    def get_release(self, release_id):
        """
        Retrieve a previously stored release.
        """
        pass

    @abc.abstractmethod
    def get_recent_releases(self, *, limit):
        """
        Return the most recent ``limit`` releases, as sorted by creation date.
        """
        pass

    def get_most_recent_release(self):
        """
        Return the most recent release, as sorted by creation date.
        """
        try:
            return self.get_recent_releases(limit=1)[0]
        except IndexError:
            raise ReleaseNotFoundError("There are no releases yet")

    @abc.abstractmethod
    def get_recent_deployments(self, *, environment=None, limit=10):
        """
        Return the most recent ``limit`` deployments in a given environment.
        """
        pass

    @abc.abstractmethod
    def add_deployment(self, *, release_id, deployment):
        """
        Store a new deployment.
        """
        pass


class MemoryReleaseStore(ReleaseStore):
    def __init__(self):
        self.cache = {}

    def describe_initialisation(self):
        return "Create in-memory release store"

    def initialise(self):
        pass

    def put_release(self, release):
        self.cache[release["release_id"]] = release

    def get_release(self, release_id):
        try:
            return self.cache[release_id]
        except KeyError:
            raise ReleaseNotFoundError(release_id)

    def get_recent_releases(self, *, limit):
        sorted_releases = sorted(
            self.cache.values(),
            key=lambda c: c["date_created"],
            reverse=True
        )
        return sorted_releases[:limit]

    def get_recent_deployments(self, *, environment=None, limit=10):
        deployments = []

        for release in self.cache.values():
            deployments.extend(release["deployments"])

        if environment:
            deployments = [d for d in deployments if d["environment"] == environment]

        return sorted(
            deployments,
            key=lambda d: d["date_created"],
            reverse=True
        )[:limit]

    def add_deployment(self, *, release_id, deployment):
        self.cache[release_id]["deployments"].append(deployment)
        self.cache[release_id]["last_date_deployed"] = deployment["date_created"]


class DynamoReleaseStore(ReleaseStore):
    def __init__(self, project_id, region_name, role_arn):
        self.project_id = project_id
        session = iam.get_session(
            session_name="ReleaseToolDynamoDbReleaseStore",
            role_arn=role_arn,
            region_name=region_name,
        )
        self.dynamodb = session.resource("dynamodb")
        self.table = self.dynamodb.Table(f"wellcome-releases-{project_id}")

    @property
    def table_name(self):
        return self.table.name

    def describe_initialisation(self):
        return f"Create DynamoDB table {self.table_name}"

    def initialise(self):
        # Attempt to load the description of the table from DynamoDB.  If this
        # fails, we know the table doesn't exist yet and we should try to
        # create it.
        try:
            self.table.load()
        except self.dynamodb.meta.client.exceptions.ResourceNotFoundException:
            self._create_table()

    def put_release(self, release):
        self.table.put_item(Item=release)

    def get_release(self, release_id):
        resp = self.table.get_item(Key={"release_id": release_id})
        try:
            return resp["Item"]
        except KeyError:
            raise ReleaseNotFoundError(release_id)

    def get_recent_releases(self, *, limit):
        # The GSI project_gsi uses date_created as a range key, so we can
        # sort by the contents of this column.
        query_resp = self.table.query(
            IndexName="project_gsi",
            ScanIndexForward=False,
            Limit=limit,
            # Note: as far as I know, this expression is a no-op -- everything
            # in a given table will have the same release ID.  We need to have
            # a KeyConditionExpression for a Query to work, and we need a Query
            # to get the sorting.
            KeyConditionExpression=Key("project_id").eq(self.project_id),
        )

        return query_resp["Items"]

    def get_recent_deployments(self, *, environment=None, limit=10):
        known_deployments = []

        params = {
            "IndexName": "deployment_gsi",
            "KeyConditionExpression": Key("project_id").eq(self.project_id),
            # Query results are always sorted by the sort key value.  Setting
            # this parameter to False means they are returned in descending order,
            # i.e. newer deployments come first.
            #
            # The sort key on this GSI is last_date_deployed.
            "ScanIndexForward": False,
        }

        items_seen = 0

        while len(known_deployments) < limit or items_seen < limit:
            resp = self.table.query(**params)

            for release in resp["Items"]:
                for deployment in release["deployments"]:
                    deployment["release_id"] = release["release_id"]

                    # Filter on environment_id - this does not play well with filter
                    # resulting in fewer than expected items (this filters on top of
                    # the limit) but it's hard to do better with the data structured
                    # as it is.
                    if environment and deployment["environment"] == environment:
                        known_deployments.append(deployment)
                    elif environment is None:
                        known_deployments.append(deployment)

            items_seen += len(resp["Items"])

            try:
                params["ExclusiveStartKey"] = resp["LastEvaluatedKey"]
            except KeyError:
                break

        known_deployments = sorted(
            known_deployments, key=lambda d: d["date_created"], reverse=True
        )

        # We then truncate the list to the limit, otherwise we might be
        # presenting an incomplete list of deployments.
        #
        # Consider:
        #
        #   - Release A
        #       deployed @ 1pm
        #       deployed @ 6pm
        #   - Release B
        #       deployed @ 2pm
        #   - Release C
        #       deployed @ 5pm
        #
        # If we requested limit=2, then DynamoDB would return the releases
        # with the two most recent "last_date_deployed" fields.  This would
        # present the following timeline:
        #
        #   - 1pm: Release A
        #   - 5pm: Release C
        #   - 6pm: Release A
        #
        # What happened to release B???
        #
        # We know we have the last N deployments with no gaps, but beyond
        # that we can't be sure.
        return known_deployments[:limit]

    def add_deployment(self, *, release_id, deployment):
        self.table.update_item(
            Key={
                'release_id': release_id
            },
            UpdateExpression="SET #deployments = list_append(#deployments, :d)",
            ExpressionAttributeNames={
                '#deployments': 'deployments',
            },
            ExpressionAttributeValues={
                ':d': [deployment],
            }
        )

        self.table.update_item(
            Key={
                'release_id': release_id
            },
            UpdateExpression="SET last_date_deployed = :d",
            ExpressionAttributeValues={
                ':d': deployment['date_created'],
            }
        )

    def _create_table(self):
        self.dynamodb.create_table(
            TableName=self.table_name,
            KeySchema=[
                {
                    'AttributeName': 'release_id',
                    'KeyType': 'HASH'
                }
            ],
            AttributeDefinitions=[
                {
                    'AttributeName': 'release_id',
                    'AttributeType': 'S'
                },
                {
                    'AttributeName': 'project_id',
                    'AttributeType': 'S'
                },
                {
                    'AttributeName': 'date_created',
                    'AttributeType': 'S'
                },
                {
                    'AttributeName': 'last_date_deployed',
                    'AttributeType': 'S'
                }
            ],
            GlobalSecondaryIndexes=[
                {
                    'IndexName': 'project_gsi',
                    'KeySchema': [
                        {
                            'AttributeName': 'project_id',
                            'KeyType': 'HASH'
                        },
                        {
                            'AttributeName': 'date_created',
                            'KeyType': 'RANGE'
                        },
                    ],
                    'Projection': {
                        'ProjectionType': 'ALL'
                    },
                    'ProvisionedThroughput': {
                        'ReadCapacityUnits': 1,
                        'WriteCapacityUnits': 1
                    }
                },
                {
                    'IndexName': 'deployment_gsi',
                    'KeySchema': [
                        {
                            'AttributeName': 'project_id',
                            'KeyType': 'HASH'
                        },
                        {
                            'AttributeName': 'last_date_deployed',
                            'KeyType': 'RANGE'
                        },
                    ],
                    'Projection': {
                        'ProjectionType': 'ALL'
                    },
                    'ProvisionedThroughput': {
                        'ReadCapacityUnits': 1,
                        'WriteCapacityUnits': 1
                    }
                }
            ],
            ProvisionedThroughput={
                'ReadCapacityUnits': 1,
                'WriteCapacityUnits': 1
            }
        )



def parse_aws_tags(tags):
    """
    When you get the tags on an AWS resource from the API, they are in the form

        [{"key": "KEY1", "value": "VALUE1"},
         {"key": "KEY2", "value": "VALUE2"},
         ...]

    This function converts them into a Python-style dict().

    """
    result = {}

    for aws_tag in tags:
        assert isinstance(aws_tag, dict)
        assert aws_tag.keys() == {"key", "value"}

        assert aws_tag["key"] not in result, f"Duplicate key in tags: {aws_tag['key']}"

        result[aws_tag["key"]] = aws_tag["value"]

    return result


class MultipleMatchingResourcesError(ValueError):
    """
    Raised if there are multiple resources matching a given set of tags.
    """


class NoMatchingResourceError(ValueError):
    """
    Raised if there is no resource matching a given set of tags.
    """
    pass


def find_unique_resource_matching_tags(resources, *, expected_tags):
    """
    Given a list of AWS resources, find the unique resource matching a given
    set of tags.

    The tags should be a Python dictionary, e.g. {"key1": "value1", "key2": "value2"}
    """
    if not expected_tags:
        raise ValueError("Cannot match against an empty set of tags")

    def _is_match(resource):
        resource_tags = parse_aws_tags(resource.get("tags", []))
        return all(
            k in resource_tags and resource_tags[k] == v
            for k, v in expected_tags.items()
        )

    matching_resources = [r for r in resources if _is_match(r)]

    if len(matching_resources) == 1:
        return matching_resources[0]
    elif not matching_resources:
        raise NoMatchingResourceError(f"Could not find any resources with tags {expected_tags}")
    else:
        raise MultipleMatchingResourcesError(
            f"Found multiple resources with tags {expected_tags}, expected one: {matching_resources}"
        )



import collections
import datetime
import functools
import uuid
import warnings

import cattr
import yaml

from . import ecr, iam, models
from .ecr import Ecr
from .ecs import Ecs
from .exceptions import ConfigError
from .release_store import DynamoReleaseStore, ReleaseNotFoundError
from .tags import parse_aws_tags

DEFAULT_REGION_NAME = "eu-west-1"


def _load(filepath):
    with open(filepath) as infile:
        return yaml.safe_load(infile)


class Projects:
    def __init__(self, project_filepath):
        self.projects = _load(project_filepath)

    def list(self):
        return list(self.projects.keys())

    def load(self, project_id, **kwargs):
        try:
            config = self.projects[project_id]
        except KeyError:
            raise ConfigError(f"No matching project {project_id} in {self.projects}")

        prepare_config(config, **kwargs)

        release_store = DynamoReleaseStore(
            project_id=project_id,
            region_name=config["region_name"],
            role_arn=config["role_arn"]
        )

        return Project(
            project_id=project_id,
            config=config,
            release_store=release_store
        )


def prepare_config(
        config,
        *,
        role_arn=None,
        region_name=None
):
    """
    Prepare the config.  Fill in overrides or defaults as necessary.
    """
    # We always want a role_arn to be set.  Read it from the initial config,
    # or raise an error if not -- there's no way to pick a sensible default.
    if role_arn:
        if ("role_arn" in config) and (config["role_arn"] != role_arn):
            warnings.warn(
                f"Preferring override role_arn {role_arn} "
                f"to role_arn in config {config['role_arn']}"
            )
            config["role_arn"] = role_arn
        elif "role_arn" not in config:
            config["role_arn"] = role_arn

    if "role_arn" not in config:
        raise ConfigError("role_arn is not set!")

    assert "role_arn" in config

    # We always want a region_name to be set.  Read it from the initial config
    # if possible, or use the override or default if not.
    if region_name and ("region_name" in config) and (config["region_name"] != region_name):
        warnings.warn(
            f"Preferring override region_name {region_name} "
            f"to region_name in config {config['region_name']}"
        )
        config["region_name"] = region_name
    elif "region_name" not in config:
        config["region_name"] = region_name or DEFAULT_REGION_NAME

    assert "region_name" in config


class Project:
    def __init__(self, project_id, config, release_store):
        self.id = project_id
        self._underlying = cattr.structure(config, models.Project)

        self.config = config
        self.config["id"] = project_id

        self.release_store = release_store
        self.release_store.initialise()

    @property
    @functools.lru_cache()
    def ecs(self):
        return self.ecs_no_cache

    @property
    def ecs_no_cache(self):
        return Ecs(
            region_name=self.region_name,
            role_arn=self.role_arn
        )

    @property
    @functools.lru_cache()
    def ecr(self):
        return Ecr(region_name=self.region_name, role_arn=self.role_arn)

    @property
    def role_arn(self):
        return self._underlying.role_arn

    @property
    def region_name(self):
        return self._underlying.region_name

    @property
    def image_repositories(self):
        return self._underlying.image_repositories

    @property
    def environment_names(self):
        return self._underlying.environments

    def _create_deployment(self, environment_id, details, description):
        return {
            "environment": environment_id,
            "date_created": datetime.datetime.utcnow().isoformat(),
            "requested_by": iam.get_underlying_role_arn(),
            "description": description,
            "details": details
        }

    def _create_release(self, description, images):
        release_id = str(uuid.uuid4())

        return {
            "release_id": release_id,
            "project_id": self.id,
            "project_name": self._underlying.name,
            "date_created": datetime.datetime.utcnow().isoformat(),
            "requested_by": iam.get_underlying_role_arn(),
            "description": description,
            "images": images,
            "deployments": []
        }

    def get_deployments(self, release_id, limit, environment_id):
        if release_id is not None:
            release = self.release_store.get_release(release_id)

            # TODO: I think the release ID is already stored on the deployments.
            # Can we remove this loop?
            for d in release["deployments"]:
                assert d["release_id"] == release_id

            deployments = release["deployments"]
        else:
            deployments = self.release_store.get_recent_deployments(
                environment=environment_id,
                limit=limit
            )

        deployments = sorted(deployments, key=lambda d: d["date_created"], reverse=True)
        return deployments[:limit]

    def get_release(self, release_id):
        if release_id == "latest":
            return self.release_store.get_most_recent_release()
        else:
            return self.release_store.get_release(release_id)

    def _get_services_by_image_id(self, release):
        """
        Generates a set of tuples (image_id, List[services])
        """
        for image_id in release["images"]:
            try:
                matched_image = self.image_repositories[image_id]
            except KeyError:
                continue

            yield (image_id, matched_image["services"])

    def get_ecs_service_arns(self, release, environment_id):
        """
        Returns a dict (image ID) -> List[service ARNs].
        """
        result = collections.defaultdict(list)

        for image_id, services in self._get_services_by_image_id(release):
            for service_id in services:
                matching_service = self.ecs.find_matching_service(
                    service_id=service_id,
                    environment_id=environment_id
                )

                try:
                    result[image_id].append(matching_service["serviceArn"])
                except TypeError:
                    assert matching_service is None, matching_service

        return dict(result)

    def get_ecs_services(self, release, environment_id, cached=True):
        def _get_service(service_id):
            # Sometimes we want not to use the service cache - eg when checking
            # whether deployments succeeded, we want a fresh copy of the services
            # information.
            if not cached:
                ecs = self.ecs_no_cache
            else:
                ecs = self.ecs

            ecs_service = ecs.find_matching_service(
                service_id=service_id,
                environment_id=environment_id
            )

            return {
                'config': {"id": service_id},
                'response': ecs_service
            }

        matched_services = {}
        for image_id, _ in release['images'].items():
            # Attempt to match deployment image id to config and override service_ids
            try:
                matched_image = self.image_repositories[image_id]
            except KeyError:
                continue

            services = matched_image.get('services', [])

            available_services = [_get_service(service_id) for service_id in services]
            available_services = [service for service in available_services if service["response"]]

            if available_services:
                matched_services[image_id] = available_services

        return matched_services

    def is_release_deployed(self, release, environment_id, verbose=False):
        """
        Checks the `deployment:label` tag on a service matches the tags
        on the tasks within those services. We check that the desiredCount
        of tasks matches the running count of tasks.
        """
        ecs_services = self.get_ecs_services(release, environment_id, cached=False)

        def printv(str):
            if verbose:
                print(str)

        is_deployed = True

        for _, services in sorted(ecs_services.items()):
            for service in services:
                service_tags = parse_aws_tags(service["response"]["tags"])
                service_arn = service["response"]["serviceArn"]
                service_deployment_label = service_tags["deployment:label"]
                desired_task_count = service["response"]["desiredCount"]
                tasks = self.ecs.list_service_tasks(service)

                for task in tasks:
                    task_tags = parse_aws_tags(task["tags"])
                    task_arn = task["taskArn"]

                    task_name = task_tags.get("deployment:service")
                    deployment_label = task_tags.get("deployment:label")

                    if deployment_label != service_deployment_label:
                        printv("")
                        printv(f"Task in {task_name} has the wrong deployment label:")
                        printv(f"Wanted:   {service_deployment_label}")
                        printv(f"Actual:   {deployment_label}")
                        printv(f"Task ARN: {task_arn}")
                        is_deployed = False

                    if task["lastStatus"] != "RUNNING":
                        printv("")
                        printv(f"Task in {task_name} has the wrong status:")
                        printv("Wanted:   RUNNING")
                        printv(f"Actual:   {task['lastStatus']}")
                        printv(f"Task ARN: {task_arn}")
                        is_deployed = False

                if len(tasks) < desired_task_count:
                    printv("")
                    printv(f"Not running the correct number of tasks in {service_arn}:")
                    printv(f"Wanted: {desired_task_count}")
                    printv(f"Actual: {len(tasks)}")
                    is_deployed = False

        return is_deployed

    def publish(self, image_id, label):
        # Create an ECR client for the correct account
        self.ecr.login()

        remote_uri, remote_tag, local_tag = self.ecr.publish_image(
            image_id=image_id,
        )

        tag_result = self.ecr.tag_image(
            image_id=image_id,
            tag=remote_tag,
            new_tag=label
        )

        return {
            'ecr_push': {
                'local_tag': local_tag,
                'remote_tag': remote_tag,
                'remote_uri': remote_uri,
            },
            'ecr_tag': tag_result
        }

    def get_images(self, from_label):
        """
        Returns a dict (image id) -> (Git ref tag).

        Note: a single Docker image may have multiple ref tags, so the ref tag
        is chosen arbitrarily.
        """
        ref_tags_resp = ecr.get_ref_tags_for_repositories(
            self.ecr._underlying.client,
            image_repositories=self.image_repositories.keys(),
            tag=from_label
        )

        # An image might have multiple ref tags if it was pushed at multiple
        # Git commits with the same code.  In this case, choose a ref arbitrarily.
        result = {}

        for image_id, ref_tags in ref_tags_resp.items():
            try:
                result[image_id] = ref_tags.pop()
            except KeyError:  # No images
                result[image_id] = None

        return result

    def _prepare_release(self, description, release_images):
        try:
            previous_release = self.release_store.get_most_recent_release()
        except ReleaseNotFoundError:
            previous_release = None

        new_release = self._create_release(
            description=description,
            images=release_images
        )

        self.release_store.put_release(new_release)

        return {"previous_release": previous_release, "new_release": new_release}

    def prepare(self, from_label, description):
        release_images = self.get_images(from_label)

        if not release_images:
            raise RuntimeError(f"No images found for {self.id}/{from_label}")

        return self._prepare_release(
            description=description,
            release_images=release_images
        )

    def update(self, release_id, service_ids, from_label):
        release = self.get_release(release_id)
        images = self.get_images(from_label)

        # Ensure all specified services are available as images
        for service_id in service_ids:
            if service_id not in images.keys():
                raise RuntimeError(f"No images found for {service_id}")

        # Filter to only specified images
        filtered_images = {k: v for k, v in images.items() if k in service_ids}

        # Merge images from specified release with those from service
        release_images = {**release['images'], **filtered_images}

        description = f"Release based on {release_id}, updating {service_ids} to {from_label}"

        return self._prepare_release(
            description=description,
            release_images=release_images
        )

    def _deploy_ecs_service(self, service, deployment_label):
        return self.ecs.redeploy_service(
            cluster_arn=service['response']['clusterArn'],
            service_arn=service['response']['serviceArn'],
            deployment_label=deployment_label
        )

    def _tag_ecr_image(self, environment_id, image_id, image_name):
        old_tag = image_name.split(":")[-1]
        new_tag = f"env.{environment_id}"

        return self.ecr.tag_image(
            image_id=image_id,
            tag=old_tag,
            new_tag=new_tag
        )

    def deploy(self, release_id, environment_id, description):
        release = self.get_release(release_id)

        if release is None:
            raise ValueError(f"No releases found {release_id}, cannot continue!")

        matched_services = self.get_ecs_services(
            release=release,
            environment_id=environment_id
        )

        # Force check for valid environment
        if environment_id not in self.environment_names:
            raise ValueError(
                f"Unknown environment. "
                f"Got {environment_id!r}, expected {self.environment_names.keys()!r}"
            )

        deployment_details = {}

        ecs_services_deployed = {}

        # Memoize service deployments to prevent multiple deployments
        def _ecs_deploy(service, deployment_label):
            if service['response']['serviceArn'] not in ecs_services_deployed:
                ecs_services_deployed[service['response']['serviceArn']] = self._deploy_ecs_service(
                    service=service,
                    deployment_label=deployment_label
                )

            return ecs_services_deployed[service['response']['serviceArn']]

        for image_id, image_name in sorted(release['images'].items()):
            tag_result = self._tag_ecr_image(
                environment_id=environment_id,
                image_id=image_id,
                image_name=image_name
            )

            if tag_result['status'] == 'noop':
                ecs_deployments = []
            else:
                ecs_deployments = [
                    _ecs_deploy(
                        service=service,
                        deployment_label=release['release_id']
                    )
                    for service in matched_services.get(image_id, [])
                ]

            deployment_details[image_id] = {
                'tag_result': tag_result,
                'ecs_deployments': ecs_deployments
            }

        deployment = self._create_deployment(environment_id, deployment_details, description)

        self.release_store.add_deployment(
            release_id=release['release_id'],
            deployment=deployment
        )

        return deployment



from abc import ABC, abstractmethod
import base64
import json
import os

from botocore.exceptions import ClientError

from . import iam
from .exceptions import EcrError
from .commands import cmd
from .git import repo_root


DEFAULT_NAMESPACE = "uk.ac.wellcome"


def create_client(*, resource, region_name, role_arn):
    session = iam.get_session(
        session_name="ReleaseToolEcr",
        role_arn=role_arn,
        region_name=region_name
    )

    return session.client(resource)


def _get_repository_name(image_id):
    return f"{DEFAULT_NAMESPACE}/{image_id}"


def get_release_image_tag(image_id):
    """
    Returns the contents of the .releases file for this image.
    """
    release_file = os.path.join(repo_root(), ".releases", image_id)
    return open(release_file).read().strip()


class AbstractEcr(ABC):
    @abstractmethod
    def base_uri(self):
        pass

    def get_image_uri(self, *, image_id, tag):
        repository_name = _get_repository_name(image_id)
        return f"{self.base_uri}/{repository_name}:{tag}"

    @abstractmethod
    def get_authorization_data(self):
        """
        Returns an "authorization token data object" that can be used to
        authenticate with the local Docker client.
        """
        pass

    def login(self):
        """
        Authenticates the local Docker client with ECR.
        """
        auth_data = self.get_authorization_data()
        auth_token = base64.b64decode(auth_data["authorizationToken"]).decode()
        username, password = auth_token.split(":")
        command = [
            "docker", "login", "--username", username, "--password", password,
            auth_data["proxyEndpoint"]
        ]

        cmd(*command)

    @abstractmethod
    def get_image_manifests_for_tag(self, *, repository_name, image_tag):
        """
        Generates a list of image manifests for a given tag.
        """
        pass

    def publish_image(self, *, image_id):
        """
        Given the ID of a local image, publish it to ECR.
        """
        local_image_tag = get_release_image_tag(image_id)
        local_image_name = f"{image_id}:{local_image_tag}"

        remote_image_tag = f"ref.{local_image_tag}"
        remote_image_name = self.get_image_uri(
            image_id=image_id,
            tag=remote_image_tag
        )

        try:
            cmd("docker", "tag", local_image_name, remote_image_name)
            cmd("docker", "push", remote_image_name)

        finally:
            cmd("docker", "rmi", remote_image_name)

        return remote_image_name, remote_image_tag, local_image_tag

    def tag_image(self, *, image_id, tag, new_tag):
        """
        Tag an image in ECR.
        """
        repository_name = _get_repository_name(image_id)

        manifests = self.get_image_manifests_for_tag(
            repository_name=repository_name,
            image_tag=tag
        )

        if len(manifests) == 0:
            raise RuntimeError(
                f"No matching images found for {repository_name}:{tag}!"
            )

        if len(manifests) > 1:
            raise RuntimeError(
                f"Multiple matching images found for {repository_name}:{tag}!"
            )

        existing_manifest = manifests[0]

        tag_operation = {
            "source": f"{repository_name}:{tag}",
            "target": f"{repository_name}:{new_tag}"
        }

        try:
            self.client.put_image(
                repositoryName=repository_name,
                imageTag=new_tag,
                imageManifest=existing_manifest
            )

            tag_operation_status = "success"
        except ClientError as e:
            # Matching tag & digest already exists (nothing to do)
            if not e.response["Error"]["Code"] == "ImageAlreadyExistsException":
                raise e
            else:
                tag_operation_status = "noop"

        tag_operation["status"] = tag_operation_status

        return tag_operation


class EcrPrivate(AbstractEcr):
    def __init__(self, *, region_name, role_arn):
        super().__init__()

        self.region_name = region_name
        self.role_arn = role_arn
        self.client = create_client(
            resource="ecr",
            region_name=region_name,
            role_arn=role_arn
        )

    @property
    def base_uri(self):
        account_id = iam.get_account_id(self.role_arn)
        return f"{account_id}.dkr.ecr.{self.region_name}.amazonaws.com"

    def get_authorization_data(self):
        resp = self.client.get_authorization_token()
        assert len(resp["authorizationData"]) == 1
        return resp["authorizationData"][0]

    def get_image_manifests_for_tag(self, *, repository_name, image_tag):
        resp = self.client.batch_get_image(
            repositoryName=repository_name,
            imageIds=[{"imageTag": image_tag}]
        )

        return [img["imageManifest"] for img in resp["images"]]


class EcrPublic(AbstractEcr):
    def __init__(self, *, gallery_id, role_arn):
        super().__init__()

        self.gallery_id = gallery_id

        self.client = create_client(
            resource="ecr-public",
            role_arn=role_arn,
            # ECR Public is a global resource that lives in us-east-1,
            # as far as I can tell.
            region_name="us-east-1"
        )

    @property
    def base_uri(self):
        return f"public.ecr.aws/{self.gallery_id}"

    def get_authorization_data(self):
        resp = self.client.get_authorization_token()
        return resp["authorizationData"]

    def get_image_manifests_for_tag(self, *, repository_name, image_tag):
        # ECR Public doesn't have an API call that returns the image manifest,
        # so we have to go out to an experimental Docker feature.
        # See https://docs.docker.com/engine/reference/commandline/manifest/
        uri = f"public.ecr.aws/{self.gallery_id}/{repository_name}"
        output = json.loads(cmd("docker", "manifest", "inspect", uri))

        return [output]


class Ecr:
    def __init__(self, *, region_name, role_arn):
        self._underlying = EcrPrivate(
            region_name=region_name,
            role_arn=role_arn
        )

    def publish_image(self, image_id):
        return self._underlying.publish_image(image_id=image_id)

    def tag_image(self, image_id, tag, new_tag):
        return self._underlying.tag_image(
            image_id=image_id,
            tag=tag,
            new_tag=new_tag
        )

    def login(self):
        self._underlying.login()


class NoSuchImageError(EcrError):
    """
    Raised when an image cannot be found.
    """
    pass


class NoRefTagError(EcrError):
    """
    Raised when an image does not have any tags starting ``ref.``.
    """
    pass


def get_ref_tags_for_image(ecr_client, *, repository_name, tag):
    """
    Returns the ref tags for the image with this tag.

    e.g. if you look for the "latest" tag, it will return the unambiguous Git ref tag(s)
    for this image.
    """
    try:
        resp = ecr_client.describe_images(
            repositoryName=repository_name,
            imageIds=[{"imageTag": tag}],
        )
    except ClientError as e:
        if not e.response["Error"]["Code"] == "ImageNotFoundException":
            raise e
        else:
            raise NoSuchImageError(
                f"Cannot find an image in {repository_name} with tag {tag}"
            )

    assert len(resp["imageDetails"]) == 1, resp
    image_details = resp["imageDetails"][0]

    tags = set(image_details["imageTags"])
    ref_tags = {t for t in tags if t.startswith("ref.")}

    if not ref_tags:
        raise NoRefTagError(
            f"No matching ref tags found for {repository_name}:{tag}!"
        )

    return ref_tags


def get_ref_tags_for_repositories(ecr_client, *, image_repositories, tag):
    """
    Returns the ref tags for all the repositories in ``image_repositories``.

    The ``image_repositories`` should be a list of repo IDs.

    Returns a dict (id) -> set(ref_tags)

    """
    result = {}

    for repo_id in image_repositories:
        repository_name = _get_repository_name(repo_id)

        try:
            ref_uri = get_ref_tags_for_image(
                ecr_client, repository_name=repository_name, tag=tag
            )
        except NoSuchImageError:
            result[repo_id] = set()
        else:
            result[repo_id] = ref_uri

    return result



import click
from distutils.version import LooseVersion
import json
import urllib

from .version import __version__ as current_version_str

current_version = LooseVersion(current_version_str)


def warn_if_not_latest_version():
    with urllib.request.urlopen("https://pypi.org/pypi/weco-deploy/json") as response:
        try:
            response_body = response.read().decode('utf-8')
            data = json.loads(response_body)
            release_versions = data["releases"].keys()
            latest_version = max([LooseVersion(v) for v in release_versions])
            latest_version_str = latest_version.vstring
            if current_version < latest_version:
                click.echo(click.style(f"You are using weco-deploy version {current_version_str}. "
                                       f"However, version {latest_version_str} is available.", fg="red"))
                click.echo(click.style("You should consider upgrading via the 'pip install --upgrade weco-deploy' "
                                       "command.", fg="yellow"))
        except urllib.error.URLError:
            # This is likely due to a network error, so ignore it
            pass
        except Exception as e:
            click.echo(click.style("Error when checking for latest version:", fg="red"))
            click.echo(e)



import boto3


def get_account_id(role_arn):
    """
    Returns the account ID for a given role ARN.
    """
    # The format of an IAM role ARN is
    #
    #     arn:partition:service:region:account:resource
    #
    # Where:
    #
    # - 'arn' is a literal string
    # - 'service' is always 'iam' for IAM resources
    # - 'region' is always blank for IAM resources
    # - 'account' is the AWS account ID with no hyphens
    #
    # See https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_identifiers.html#identifiers-arns
    try:
        arn, _, service, region, account, _ = role_arn.split(":")
    except ValueError:
        raise ValueError(f"Is this a valid AWS ARN? {role_arn}")

    if arn != "arn":
        raise ValueError(f"Is this a valid AWS ARN? {role_arn}")

    if service != "iam" or region != "" or not account.isnumeric():
        raise ValueError(f"Is this an IAM role ARN? {role_arn}")

    return account


def get_underlying_role_arn():
    """
    Returns the original role ARN.

    e.g. at Wellcome we have a base role, but then we assume roles into different
    accounts.  This returns the ARN of the base role.
    """
    client = boto3.client('sts')
    return client.get_caller_identity()["Arn"]


def get_session(session_name, role_arn, region_name):
    client = boto3.client('sts')
    response = client.assume_role(
        RoleArn=role_arn,
        RoleSessionName=session_name
    )

    return boto3.session.Session(
        aws_access_key_id=response['Credentials']['AccessKeyId'],
        aws_secret_access_key=response['Credentials']['SecretAccessKey'],
        aws_session_token=response['Credentials']['SessionToken'],
        region_name=region_name
    )



import collections
import itertools

from .exceptions import ConfigError


def chunked_iterable(iterable, *, size):
    """
    Generate the entries in ``iterable`` in chunks of size ``size``.

    e.g. chunked_iterable([1, 2, 3, 4, 5], 2) -> [1, 2], [3, 4], [5]

    Taken from https://alexwlchan.net/2018/12/iterating-in-fixed-size-chunks/
    """
    it = iter(iterable)
    while True:
        chunk = tuple(itertools.islice(it, size))
        if not chunk:
            break
        yield chunk


def convert_identified_list_to_dict(values):
    """
    Given a list of objects with a .id parameter, convert them to a dict
    keyed with the .id.
    """
    id_counts = collections.Counter(v.id for v in values)
    duplicate_ids = {id for id, count in id_counts.items() if count > 1}
    if duplicate_ids:
        raise ConfigError("Duplicate IDs: %s" % ", ".join(duplicate_ids))

    return {v.id: v for v in values}



import typing

import attr
import cattr
import yaml

from .iterators import convert_identified_list_to_dict


@attr.s
class Environment:
    id = attr.ib()
    name = attr.ib()


@attr.s
class Service:
    id = attr.ib()


@attr.s
class ImageRepository:
    id = attr.ib()
    services: typing.List[Service] = attr.ib(
        factory=list, converter=convert_identified_list_to_dict
    )


@attr.s
class Project:
    name = attr.ib()
    role_arn = attr.ib()

    environments: typing.List[Environment] = attr.ib(
        factory=list, converter=convert_identified_list_to_dict
    )
    image_repositories: typing.List[ImageRepository] = attr.ib(
        factory=list, converter=convert_identified_list_to_dict
    )

    aws_region_name = attr.ib(default="eu-west-1")

    @property
    def region_name(self):
        return self.aws_region_name


class ProjectList:
    @classmethod
    def from_text(cls, yaml_text):
        data = yaml.safe_load(yaml_text)
        return cattr.structure(data, typing.Dict[str, Project])

    @classmethod
    def from_path(cls, path):
        with open(path) as infile:
            yaml_text = infile.read()
        return ProjectList.from_text(yaml_text=yaml_text)
